#include <QtGui>
#include "Correspondence.h"
#include "math.h"
#include "ImageProcessing.h"
#include <iostream>
#include "Vector.h"
#include "FlowInterpolation.h"
#include "QVideoProcessing.h"
#include "VideoLabeling.h"


using namespace std;

int Correspondence::m_Left=0;
int Correspondence::m_Top=0;
int Correspondence::m_Width=0;
int Correspondence::m_Height=0;

Correspondence::Correspondence(void): QObject()
{
	pointIndex=0;
	motionType=Translation;
	affinePara.allocate(2,3);
	homographyPara.allocate(8);
	IsMotionTypeSpecified=false;
	IsFlowInterpolated=true;
	m_Lambda=1;
}

Correspondence::~Correspondence(void)
{
}

Correspondence::Correspondence(const Correspondence& other)
{
	copyData(other);
}

Correspondence& Correspondence::operator =(const Correspondence& other)
{
	copyData(other);
	return *this;
}

void Correspondence::copyData(const Correspondence& other)
{
	pointList1=other.pointList1;
	pointList2=other.pointList2;
	uncertaintyList=other.uncertaintyList;
	candidateList=other.candidateList;
	densityList=other.densityList;
	invCovList=other.invCovList;
	meanList=other.meanList;
	pointIndex=other.pointIndex;
	motionType=other.motionType;
	IsMotionTypeSpecified=other.IsMotionTypeSpecified;
	transPara=other.transPara;
	affinePara=other.affinePara;
	homographyPara=other.homographyPara;
	flow=other.flow;
	IsFlowInterpolated=other.IsFlowInterpolated;
	m_Lambda=other.m_Lambda;
}



void Correspondence::clear()
{
	pointList1.clear();
	pointList2.clear();
	
	uncertaintyList.clear();
	candidateList.clear();
	densityList.clear();

	invCovList.clear();
	meanList.clear();

	pointIndex=0;
	motionType=Translation;
	IsMotionTypeSpecified=false;
	IsFlowInterpolated=true;
	emit motionTypeChanged(0);
}

void Correspondence::specifyMotionType(int index)
{
	switch(index){
		case 0:
			motionType=Translation;
			break;
		case 1:
			motionType=Affine;
			break;
		case 2:
			motionType=Homography;
			break;
		case 3:
			motionType=PiecewiseSmooth;
			break;
	}
	if(pointList1.size()>0)
		IsMotionTypeSpecified=true;
}

int Correspondence::getMotionType() const
{
	switch(motionType){
		case Translation:
			return 0;
		case Affine:
			return 1;
		case Homography:
			return 2;
		case PiecewiseSmooth:
			return 3;
	}
	return 0;
}

//---------------------------------------------------------------------------------------------------------------
// function to draw the correspondence
//---------------------------------------------------------------------------------------------------------------
void Correspondence::DrawCorresponance(QPainter &painter, const QPoint& ScreenOffset, double ScaleFactor)
{
	if(pointList1.size()==0)
		return;
	// draw the lines
	QPen pen;
	pen.setColor(QColor(255,128,0));
	pen.setWidthF(sqrt(ScaleFactor));
	painter.setPen(pen);
	for(int i=0;i<pointList1.size();i++)
		painter.drawLine(ScreenOffset+pointList1[i]*ScaleFactor,ScreenOffset+pointList2[i]*ScaleFactor);

	// draw the start points
	//pen.setColor(Qt::red);
	//pen.setWidthF(1);
	painter.setPen(Qt::red);
	double wsize=2*sqrt(ScaleFactor);
	for(int i=0;i<pointList1.size();i++)
	{
		QPointF point(ScreenOffset+pointList1[i]*ScaleFactor);
		painter.drawEllipse(QRectF(point.x()-wsize,point.y()-wsize,wsize*2+1,wsize*2+1));
	}
	
	for(int i=0;i<pointList1.size();i++)
	{
		QPointF point(ScreenOffset+pointList2[i]*ScaleFactor);
		if(uncertaintyList[i]==false)
			painter.setPen(Qt::green);
		else
			painter.setPen(Qt::cyan);
		painter.drawEllipse(QRectF(point.x()-wsize,point.y()-wsize,wsize*2+1,wsize*2+1));
	}
}

void Correspondence::addCorrespondence(const QPointF &point1, const QPointF &point2)
{
	pointList1.append(point1);
	pointList2.append(point2);
	uncertaintyList.append(false);
	QVector<QPointF> candidates;
	QVector<double> density;
	candidateList.append(candidates);
	densityList.append(density);

	Matrix M(2,2);
	M.identity(2);
	M*=6;
	Vector b(2);
	b[0]=point2.x();
	b[1]=point2.y();
	invCovList.append(M);
	meanList.append(b);

	pointIndex=pointList1.size()-1;
}

void Correspondence::addCorrespondence(const QPointF& point1,const SortQueue<Point> & sort)
{
	pointList1.append(point1);
	pointList2.append(QPointF(sort[0].x(),sort[0].y()));
	uncertaintyList.append(true);

	QVector<QPointF> candidates(sort.num());
	QVector<double> density(sort.num());
	for(int i=0;i<sort.num();i++)
	{
		candidates[i].setX(sort[i].x());
		candidates[i].setY(sort[i].y());
		double temp=sort.score(i)/sort.score(0);
		density[i]=exp(-temp*temp);
	}
	candidateList.append(candidates);
	densityList.append(density);

	Matrix M(2,2);
	Vector b(2);
	computeInvCov(M,b,candidates,density);
	M.printMatrix();
	cout<<endl;
	invCovList.append(M);
	meanList.append(b);

	pointIndex=pointList1.size()-1;
}

void Correspondence::computeInvCov(Matrix &invCov, Vector &mean, const QVector<QPointF> &candidates, const QVector<double> density)
{
	// compute the total weight
	double sumWeight=0;
	for(int i=0;i<density.size();i++)
		sumWeight+=density[i];

	// first compute mean;
	if(!mean.matchDimension(2))
		mean.allocate(2);
	mean.reset();
	for(int i=0;i<candidates.size();i++)
	{
		mean[0]+=candidates[i].x()*density[i];
		mean[1]+=candidates[i].y()*density[i];
	}
	mean/=sumWeight;

	// compute the covariance matrix
	Matrix Cov(2,2);
	Cov.reset();
	for(int i=0;i<candidates.size();i++)
	{
		double x=candidates[i].x()-mean[0];
		double y=candidates[i].y()-mean[1];
		Cov.data()[0]+=x*x*density[i];
		Cov.data()[1]+=x*y*density[i];
		Cov.data()[2]+=x*y*density[i];
		Cov.data()[3]+=y*y*density[i];
	}
	Cov/=sumWeight;

	// compute inverse covariance matrix
	double det=Cov.data()[0]*Cov.data()[3]-Cov.data()[1]*Cov.data()[2];
	if(!invCov.matchDimension(2,2))
		invCov.allocate(2,2);
	invCov.data()[0]=Cov.data()[3]/det;
	invCov.data()[1]=-Cov.data()[1]/det;
	invCov.data()[2]=-Cov.data()[2]/det;
	invCov.data()[3]=Cov.data()[0]/det;

	// display the covariance matrix and the mean
	//mean.printVector();
	//cout<<"\n";
	//invCov.printMatrix();
	//cout<<"\n";
}

void Correspondence::updateSecond(const QPointF& point2)
{
	pointList2[pointIndex]=point2;
}

void Correspondence::applyTransform(const QImage &image2, QImage &warpImage2,double scaleFactor)
{
	// for now the images are blowed by 2 for better visualization
	if(pointList1.size()==0)
	{
		warpImage2=image2;
		//cout<<"no points!"<<endl;
		return;
	}
	
	//estimateMotion(scaleFactor);
//	scaleMotion(scaleFactor);
	
	if(warpImage2.width()!=image2.width() || warpImage2.height()!=image2.height())
		warpImage2=image2;

	// bilinear interpolation to generate the warped image
	const unsigned char **pScanLines;
	pScanLines=new const unsigned char*[image2.height()];
	for(int i=0;i<image2.height();i++)
		pScanLines[i]=image2.scanLine(i);
	int xx,yy,u,v;
	double x,y;
	double dx,dy,s;
	int result[3];
	int width=image2.width();
	int height=image2.height();
	double denominator;

	for(int i=0;i<height;i++)
	{
		unsigned char* scanline=warpImage2.scanLine(i);
		memset(scanline,255,width*4);
		for(int j=0;j<width;j++)
		{
			applyMotion(x,y,j,i);
			xx=x;
			yy=y;
                        dx=__max(__min(x-xx,1.0),0.0);
                        dy=__max(__min(y-yy,1.0),0.0);
			for(int l=0;l<3;l++)
				result[l]=0;
			for(int m=0;m<=1;m++)
				for(int n=0;n<=1;n++)
				{
					u=ImageProcessing::EnforceRange(xx+m,width);
					v=ImageProcessing::EnforceRange(yy+n,height);
					s=fabs(1-m-dx)*fabs(1-n-dy);
					for(int l=0;l<3;l++)
						result[l]+=pScanLines[v][u*4+l]*s;
				}
			for(int l=0;l<3;l++)
				scanline[j*4+l]=__min(__max(result[l],0),255);
		}
	}
	delete pScanLines;
}

void Correspondence::applyMotion(double& x,double& y,int j,int i)
{
	double denominator;
	switch(motionType){
		case Translation:
			x=j+transPara.data()[0];
			y=i+transPara.data()[1];
			break;
		case Affine:
			x=j*affinePara.data()[0]+i*affinePara.data()[1]+affinePara.data()[2];
			y=j*affinePara.data()[3]+i*affinePara.data()[4]+affinePara.data()[5];
			break;
		case Homography:
			denominator=j*homographyPara.data()[6]+i*homographyPara.data()[7]+1;
			x=(j*homographyPara.data()[0]+i*homographyPara.data()[1]+homographyPara.data()[2])/denominator;
			y=(j*homographyPara.data()[3]+i*homographyPara.data()[4]+homographyPara.data()[5])/denominator;
			break;
		case PiecewiseSmooth:
			int offset=(i*flow.width()+j)*2;
			x=j+flow.data()[offset];
			y=i+flow.data()[offset+1];
			break;
	}
}

void Correspondence::generateFlowField(DImage &vx, DImage &vy, int left, int top, int right, int bottom)
{
	int width=right-left+1;
	int height=bottom-top+1;
	if(!vx.matchDimension(width,height,1))
		vx.allocate(width,height);
	if(!vy.matchDimension(width,height,1))
		vy.allocate(width,height);
	if(pointList1.size()==0)
	{
		vx.reset();
		vy.reset();
		return;
	}
	if(motionType==PiecewiseSmooth && (flow.width()!=vx.width() || flow.height()!=vy.height()))
	{
		setFlowDimension(left,top,width,height);
		estimatePeicewiseSmooth();
	}

	double x,y;
	for(int i=0;i<height;i++)
		for(int j=0;j<width;j++)
		{
			if(motionType!=PiecewiseSmooth)
			{
				applyMotion(x,y,j+left,i+top);
				vx.data()[i*width+j]=x-(j+left);
				vy.data()[i*width+j]=y-(i+top);
			}
			else
			{
				int offset=i*width+j;
				vx.data()[offset]=flow.data()[offset*2];
				vy.data()[offset]=flow.data()[offset*2+1];
			}
		}
}

void Correspondence::estimateMotion()
{
	if(IsMotionTypeSpecified==false)
	{
		if(pointList1.size()<5)
		{
			motionType=Translation;
			emit motionTypeChanged(0);
		}
		else if(pointList1.size()<10)
		{
			motionType=Affine;
			emit motionTypeChanged(1);
		}
		else
		{
			motionType=Homography;
			emit motionTypeChanged(2);
		}
		IsMotionTypeSpecified=false;
	}

	switch(motionType){
		case Translation:
			estimateTranslationUncertain();
			break;
		case Affine:
			estimateAffineUncertain();
			break;
		case Homography:
			//estimateAffine(scaleFactor);
			estimateHomographyUncertain();
			break;
		case PiecewiseSmooth:
			if(IsFlowInterpolated==false)
				estimatePeicewiseSmooth();
			break;
	}
}

/*
void Correspondence::scaleMotion(double scaleFactor)
{
	switch(motionType){
		case Translation:
			scaledTransPara=transPara*scaleFactor;
			break;
		case Affine:
			scaledAffinePara=affinePara;
			scaledAffinePara.data()[2]*=scaleFactor;
			scaledAffinePara.data()[5]*=scaleFactor;
			break;
		case Homography:
			scaledHomographyPara=homographyPara;
			scaledHomographyPara.data()[2]*=scaleFactor;
			scaledHomographyPara.data()[5]*=scaleFactor;
			scaledHomographyPara.data()[6]/=scaleFactor;
			scaledHomographyPara.data()[7]/=scaleFactor;
			break;
		case PiecewiseSmooth:
			flow.imresize(scaledFlow,imWidth*scaleFactor,imHeight*scaleFactor);
			break;
	}
}//*/

//-------------------------------------------------------------------------------------------------
// function to estimate translation
//-------------------------------------------------------------------------------------------------
void Correspondence::estimateTranslation()
{
	transPara.reset();
	for(int i=0;i<pointList1.size();i++)
	{
		QPointF point=pointList2[i]-pointList1[i];
		transPara[0]+=point.x();
		transPara[1]+=point.y();
	}
	transPara/=pointList1.size();
}

void Correspondence::estimateTranslationUncertain()
{
	if(!transPara.matchDimension(2))
		transPara.allocate(2);
	bool IsAllCertain=true;
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i]==true)
		{
			IsAllCertain=false;
			break;
		}
	if(IsAllCertain)
	{
		estimateTranslation();
		return;
	}

	QList<Matrix> InvCovList(invCovList);
	QList<Vector> MeanList(meanList);

	int nIterations=8;
	for(int k=0;k<nIterations;k++)
	{
		Matrix M(2,2);
		Vector b(2);
		M.reset();b.reset();

		for(int i=0;i<InvCovList.size();i++)
		{
			M+=InvCovList[i];
			Vector temp(2);
			temp[0]=MeanList[i][0]-pointList1[i].x();
			temp[1]=MeanList[i][1]-pointList1[i].y();
			b+=InvCovList[i]*temp;
		}

		// solve the system
		M.ConjugateGradient(transPara,b);

		transPara.printVector();
		cout<<endl;

		// update the weight
		if(k==nIterations-1)
			break;

		for(int i=0;i<InvCovList.size();i++)
		{
			if(uncertaintyList[i]==false)
				continue;
			double x2,y2;
			x2=pointList1[i].x()+transPara[0];
			y2=pointList1[i].y()+transPara[1];
			QVector<double> updatedDensity(densityList[i].size());
			for(int j=0;j<densityList[i].size();j++) // update the weight according to the distance
			{
				double temp1=x2-candidateList[i][j].x();
				double temp2=y2-candidateList[i][j].y();
				updatedDensity[j]=exp(-(temp1*temp1+temp2*temp2)/60)*densityList[i][j];
			}
			computeInvCov(InvCovList[i],MeanList[i],candidateList[i],updatedDensity);
		}
	}
	// apply the transform to the uncertain points
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i])
		{
			pointList2[i].setX(pointList1[i].x()+transPara[0]);
			pointList2[i].setY(pointList1[i].y()+transPara[1]);
		}
}

//-------------------------------------------------------------------------------------------------
// function to estimate affine motion
//-------------------------------------------------------------------------------------------------
void Correspondence::estimateAffine()
{
	if(!affinePara.matchDimension(2,3))
		affinePara.allocate(2,3);
	Matrix A(pointList1.size(),3);
	for(int i=0;i<pointList1.size();i++)
	{
		A.data()[i*3]=pointList1[i].x();
		A.data()[i*3+1]=pointList1[i].y();
		A.data()[i*3+2]=1;
	}
	Vector b(pointList1.size()),para;
	for(int i=0;i<pointList1.size();i++)
		b[i]=pointList2[i].x();
	A.SolveLinearSystem(para,b);
	for(int i=0;i<3;i++)
		affinePara.data()[i]=para[i];

	for(int i=0;i<pointList1.size();i++)
		b[i]=pointList2[i].y();
	A.SolveLinearSystem(para,b);
	for(int i=0;i<3;i++)
		affinePara.data()[i+3]=para[i];

	affinePara.printMatrix();
	cout<<endl;
}

void Correspondence::estimateAffineUncertain()
{
	if(!affinePara.matchDimension(2,3))
		affinePara.allocate(2,3);
	bool IsAllCertain=true;
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i]==true)
		{
			IsAllCertain=false;
			break;
		}
	if(IsAllCertain)
	{
		estimateAffine();
		return;
	}

	QList<Matrix> InvCovList(invCovList);
	QList<Vector> MeanList(meanList);

	int nIterations=30;
	double sigma=6;
	for(int k=0;k<nIterations;k++)
	{
		Matrix M(6,6);
		Vector b(6);
		Matrix X(6,2),XT(2,6);
		M.reset();b.reset();
		X.reset();XT.reset();

		for(int i=0;i<InvCovList.size();i++)
		{
			XT.data()[0]=pointList1[i].x();
			XT.data()[1]=pointList1[i].y();
			XT.data()[2]=1;
			XT.data()[9]=pointList1[i].x();
			XT.data()[10]=pointList1[i].y();
			XT.data()[11]=1;
			XT.transpose(X);

			M+=X*InvCovList[i]*XT;
			b+=X*InvCovList[i]*MeanList[i];
		}

		// solve the system
		Vector tempPara(6);
		M.ConjugateGradient(tempPara,b);
		for(int i=0;i<6;i++)
			affinePara.data()[i]=tempPara[i];

		affinePara.printMatrix();
		cout<<endl;

		// update the weight
		if(k==nIterations-1)
			break;

		for(int i=0;i<InvCovList.size();i++)
		{
			if(uncertaintyList[i]==false)
				continue;
			double x2,y2;
			x2=pointList1[i].x()*affinePara.data()[0]+pointList1[i].y()*affinePara.data()[1]+affinePara.data()[2];
			y2=pointList1[i].x()*affinePara.data()[3]+pointList1[i].y()*affinePara.data()[4]+affinePara.data()[5];
			QVector<double> updatedDensity(densityList[i].size());
			for(int j=0;j<densityList[i].size();j++) // update the weight according to the distance
			{
				double temp1=x2-candidateList[i][j].x();
				double temp2=y2-candidateList[i][j].y();
				updatedDensity[j]=exp(-(temp1*temp1+temp2*temp2)/(2*sigma*sigma))*densityList[i][j];
			}
			computeInvCov(InvCovList[i],MeanList[i],candidateList[i],updatedDensity);
		}
		//sigma*=0.99;
	}
	// apply the transform to the uncertain points
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i])
		{
			double x2,y2;
			x2=pointList1[i].x()*affinePara.data()[0]+pointList1[i].y()*affinePara.data()[1]+affinePara.data()[2];
			y2=pointList1[i].x()*affinePara.data()[3]+pointList1[i].y()*affinePara.data()[4]+affinePara.data()[5];
			pointList2[i].setX(x2);
			pointList2[i].setY(y2);
		}
	cout<<endl;
}

//-------------------------------------------------------------------------------------------------
// function to estimate homography
//-------------------------------------------------------------------------------------------------
void Correspondence::estimateHomography()
{
	if(!homographyPara.matchDimension(8))
		homographyPara.allocate(8);
	Matrix A(pointList1.size()*2,8);
	Vector b(pointList1.size()*2);
	A.reset();
	b.reset();
	for(int i=0;i<pointList1.size()*2;i++)
	{
		int index=i/2;
		if(i%2==0) // x
		{
			A.data()[i*8]=pointList1[index].x();
			A.data()[i*8+1]=pointList1[index].y();
			A.data()[i*8+2]=1;
			A.data()[i*8+6]=-pointList1[index].x()*pointList2[index].x();
			A.data()[i*8+7]=-pointList1[index].y()*pointList2[index].x();
			b.data()[i]=pointList2[index].x();
		}
		else //y
		{
			A.data()[i*8+3]=pointList1[index].x();
			A.data()[i*8+4]=pointList1[index].y();
			A.data()[i*8+5]=1;
			A.data()[i*8+6]=-pointList1[index].x()*pointList2[index].y();
			A.data()[i*8+7]=-pointList1[index].y()*pointList2[index].y();
			b.data()[i]=pointList2[index].y();
		}
	}
	Matrix::enableDispInfo(true);
	A.SolveLinearSystem(homographyPara,b);
	Matrix::enableDispInfo(false);
	homographyPara.printVector();
	// test whether it is a good match
	for(int i=0;i<pointList1.size();i++)
	{
		double x0,y0,x,y;
		x0=pointList1[i].x();
		y0=pointList1[i].y();
		double denominator=x0*homographyPara.data()[6]+y0*homographyPara.data()[7]+1;
		x=(x0*homographyPara.data()[0]+y0*homographyPara.data()[1]+homographyPara.data()[2])/denominator;
		y=(x0*homographyPara.data()[3]+y0*homographyPara.data()[4]+homographyPara.data()[5])/denominator;
		double x2=pointList2[i].x(),y2=pointList2[i].y();
		printf("No. %d error=%f  denominator=%f\n",i,sqrt((x2-x)*(x2-x)+(y2-y)*(y2-y)),denominator);
	}
	printf("\n");
}

void Correspondence::estimateHomographyUncertain()
{
	if(!homographyPara.matchDimension(8))
		homographyPara.allocate(8);
	bool IsAllCertain=true;
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i]==true)
		{
			IsAllCertain=false;
			break;
		}
	if(IsAllCertain)
	{
		estimateHomography();
		return;
	}

	QList<Matrix> InvCovList(invCovList);
	QList<Vector> MeanList(meanList);

	int nIterations=30;
	double sigma=6;
	for(int k=0;k<nIterations;k++)
	{
		Matrix M(8,8);
		Vector b(8);
		Matrix X(8,2),XT(2,8);
		M.reset();b.reset();
		X.reset();XT.reset();

		for(int i=0;i<InvCovList.size();i++)
		{
			XT.data()[0]=pointList1[i].x();
			XT.data()[1]=pointList1[i].y();
			XT.data()[2]=1;
			XT.data()[6]=-pointList1[i].x()*MeanList[i][0];
			XT.data()[7]=-pointList1[i].y()*MeanList[i][0];
			XT.data()[11]=pointList1[i].x();
			XT.data()[12]=pointList1[i].y();
			XT.data()[13]=1;
			XT.data()[14]=-pointList1[i].x()*MeanList[i][1];
			XT.data()[15]=-pointList1[i].y()*MeanList[i][1];
			XT.transpose(X);

			M+=X*InvCovList[i]*XT;
			b+=X*InvCovList[i]*MeanList[i];
		}

		// solve the system
		
		M.ConjugateGradient(homographyPara,b);

		homographyPara.printVector();

		// update the weight
		if(k==nIterations-1)
			break;

		for(int i=0;i<InvCovList.size();i++)
		{
			if(uncertaintyList[i]==false)
				continue;
			double x1,y1,x2,y2,denominator;
			x1=pointList1[i].x();
			y1=pointList1[i].y();
			denominator=x1*homographyPara[6]+y1*homographyPara[7]+1;
			x2=(x1*homographyPara[0]+y1*homographyPara[1]+homographyPara[2])/denominator;
			y2=(x1*homographyPara[3]+y1*homographyPara[4]+homographyPara[5])/denominator;

			QVector<double> updatedDensity(densityList[i].size());
			for(int j=0;j<densityList[i].size();j++) // update the weight according to the distance
			{
				double temp1=x2-candidateList[i][j].x();
				double temp2=y2-candidateList[i][j].y();
				updatedDensity[j]=exp(-(temp1*temp1+temp2*temp2)/(2*sigma*sigma))*densityList[i][j];
			}
			computeInvCov(InvCovList[i],MeanList[i],candidateList[i],updatedDensity);
		}
		//sigma*=0.99;
	}
	cout<<endl;

	// apply the transform to the uncertain points
	for(int i=0;i<uncertaintyList.size();i++)
		if(uncertaintyList[i])
		{
			double x1,y1,x2,y2,denominator;
			x1=pointList1[i].x();
			y1=pointList1[i].y();
			denominator=x1*homographyPara[6]+y1*homographyPara[7]+1;
			x2=(x1*homographyPara[0]+y1*homographyPara[1]+homographyPara[2])/denominator;
			y2=(x1*homographyPara[3]+y1*homographyPara[4]+homographyPara[5])/denominator;
			pointList2[i].setX(x2);
			pointList2[i].setY(y2);
		}
}

void Correspondence::estimatePeicewiseSmooth()
{
	FlowInterpolation flowInterp;
	flowInterp.Initialize(m_Width,m_Height);
	flowInterp.setLevel(10);
	QList<Vector> xList;
	QList<Vector> MeanList(meanList);
	QList<Matrix> InvCovList(invCovList);
	for(int i=0;i<pointList1.size();i++)
	{
		Vector vect(2);
		vect[0]=(int)pointList1[i].x();
		vect[1]=(int)pointList1[i].y();
		MeanList[i]=meanList[i]-vect;
		vect[0]-=m_Left;
		vect[1]-=m_Top;
		xList.append(vect);
		
		//if(uncertaintyList[i])
		//	InvCovList[i]*=20;
		//InvCovList[i].printMatrix();
		//cout<<endl;
	}
	cout<<"Lambda: "<<m_Lambda<<endl;
	//cout<<"Program goes here"<<endl;
	flowInterp.Interpolate(xList,InvCovList,MeanList,1/m_Lambda,__min(m_Width,m_Height)*4);
	flow=flowInterp.flow();
	flowInterp.dispAccuracy();
	IsFlowInterpolated=true;
}


double Correspondence::norm(const QPointF& point)
{
	return sqrt(point.x()*point.x()+point.y()*point.y());
}

bool Correspondence::selectPoint(const QPointF &query,double scaleFactor)
{
	if(pointList1.size()==0)
		return false;
	double MinDistance=10000;
	int index=-1;
	for(int i=0;i<pointList1.size();i++)
	{
		double temp=norm(pointList1[i]-query);
		if(temp<MinDistance)
		{
			MinDistance=temp;
			index=i;
		}
	}
	if(MinDistance<=sqrt(scaleFactor)*1.1)
	{
		pointIndex=index;
		return true;
	}
	else
		return false;
}

void Correspondence::removePoint()
{
	pointList1.removeAt(pointIndex);
	pointList2.removeAt(pointIndex);
	uncertaintyList.removeAt(pointIndex);
	candidateList.removeAt(pointIndex);
	densityList.removeAt(pointIndex);
	invCovList.removeAt(pointIndex);
	meanList.removeAt(pointIndex);
}

//------------------------------------------------------------------------------------------------------
// function to crop an image
//------------------------------------------------------------------------------------------------------
void Correspondence::getPatch(QImage &patch, const QImage &srcImage, double x0, double y0, int wsize,double scale)
{
	int winlength=wsize*2+1;
	if(patch.width()!=winlength || patch.height()!=winlength)
	{
		QImage temp(winlength,winlength,QImage::Format_RGB32);
		patch=temp;
	}
	// bilinear interpolation to generate the warped image
	const unsigned char **pScanLines;
	pScanLines=new const unsigned char*[srcImage.height()];
	for(int i=0;i<srcImage.height();i++)
		pScanLines[i]=srcImage.scanLine(i);
	int xx,yy,u,v;
	double x,y;
	double dx,dy,s;
	int result[3];
	int width=srcImage.width();
	int height=srcImage.height();
	for(int i=0;i<winlength;i++)
	{
		unsigned char* scanline=patch.scanLine(i);
		memset(scanline,255,winlength*4);
		for(int j=0;j<winlength;j++)
		{
			x=x0+(j-wsize)*scale;
			y=y0+(i-wsize)*scale;
			xx=x;
			yy=y;
			if(xx<0 || yy<0 || xx>width-1 || yy>height-1)
			{
				for(int l=0;l<3;l++)
					scanline[j*4+l]=0;
				continue;
			}
                        dx=__max(__min(x-xx,1.0),0.0);
                        dy=__max(__min(y-yy,1.0),0.0);
			for(int l=0;l<3;l++)
				result[l]=0;
			for(int m=0;m<=1;m++)
				for(int n=0;n<=1;n++)
				{
					u=ImageProcessing::EnforceRange(xx+m,width);
					v=ImageProcessing::EnforceRange(yy+n,height);
					s=fabs(1-m-dx)*fabs(1-n-dy);
					for(int l=0;l<3;l++)
						result[l]+=pScanLines[v][u*4+l]*s;
				}
			for(int l=0;l<3;l++)
				scanline[j*4+l]=__min(__max(result[l],0),255);
		}
	}
	delete pScanLines;
}

//---------------------------------------------------------------------------------------------------------------
// function to write correspondence
//---------------------------------------------------------------------------------------------------------------
bool Correspondence::writeCorrespondence(QFile &file) const
{
	// first write the correspondence
	if(!writeQPointFList(file,pointList1))
		return false;
	if(!writeQPointFList(file,pointList2))
		return false;
	if(!QVideoProcessing::writeQList(file,uncertaintyList))
		return false;
	
	int nPoints=candidateList.size();
	file.write((char *)&nPoints,sizeof(int));

	for(int i=0;i<nPoints;i++)
	{
		if(!writeQPointFVector(file,candidateList[i]))
			return false;
		if(!QVideoProcessing::writeQVector(file,densityList[i]))
			return false;
		if(!invCovList[i].writeMatrix(file))
			return false;
		if(!meanList[i].writeVector(file))
			return false;
	}

	file.write((char *)&pointIndex,sizeof(int));
	file.write((char *)&motionType,sizeof(MotionType));
	file.write((char *)&IsMotionTypeSpecified,sizeof(bool));

	// output motion parameter
	if(!transPara.writeVector(file))
		return false;
	if(!affinePara.writeMatrix(file))
		return false;
	if(!homographyPara.writeVector(file))
		return false;

	file.write((char *)&m_Lambda,sizeof(double));
	if(!flow.writeImage(file))
		return false;

	file.write((char *)&IsFlowInterpolated,sizeof(int));

	return true;
}


bool Correspondence::readCorrespondence(QFile &file)
{
	clear();

	// first read the correspondence
	if(!readQPointFList(file,pointList1))
		return false;
	if(!readQPointFList(file,pointList2))
		return false;
	if(!QVideoProcessing::readQList(file,uncertaintyList))
		return false;
	
	int nPoints;
	file.read((char *)&nPoints,sizeof(int));
	if(nPoints<0)
		return false;
	
        candidateList=QList<QVector<QPointF> >::fromVector(QVector<QVector<QPointF> >(nPoints));
        densityList=QList<QVector<double> >::fromVector(QVector<QVector<double> >(nPoints));
	invCovList=QList<Matrix>::fromVector(QVector<Matrix>(nPoints));
	meanList=QList<Vector>::fromVector(QVector<Vector>(nPoints));

	for(int i=0;i<nPoints;i++)
	{
		if(!readQPointFVector(file,candidateList[i]))
			return false;
		if(!QVideoProcessing::readQVector(file,densityList[i]))
			return false;
		if(!invCovList[i].readMatrix(file))
			return false;
		if(!meanList[i].readVector(file))
			return false;
	}

	file.read((char *)&pointIndex,sizeof(int));
	file.read((char *)&motionType,sizeof(MotionType));
	file.read((char *)&IsMotionTypeSpecified,sizeof(bool));

	// read motion parameter
	if(!transPara.readVector(file))
		return false;
	if(!affinePara.readMatrix(file))
		return false;
	if(!homographyPara.readVector(file))
		return false;

	if(VideoLabeling::curVersion()>1.19) // for version equal to or greater than 1.20
		file.read((char *)&m_Lambda,sizeof(double));
	else
		m_Lambda=1;

	if(!flow.readImage(file))
		return false;

	file.read((char *)&IsFlowInterpolated,sizeof(int));

	return true;
}


// these are static function to help write and read QList and QVector objects
bool Correspondence::writeQPointFList(QFile& file,const QList<QPointF> &pointList)
{
	int num=pointList.size();
	file.write((char *)&num,sizeof(int));
	double *pBuffer=new double[num*2];
	for(int i=0;i<num;i++)
	{
		pBuffer[i*2]=pointList[i].x();
		pBuffer[i*2+1]=pointList[i].y();
	}
	if(file.write((char *)pBuffer,sizeof(double)*num*2)!=sizeof(double)*num*2)
	{
		delete pBuffer;
		return false;
	}
	delete pBuffer;
	return true;
}

bool Correspondence::readQPointFList(QFile &file, QList<QPointF> &pointList)
{
	int num;
	file.read((char *)&num,sizeof(int));
	if(num<0)
		return false;
	double* pBuffer=new double[num*2];
	if(file.read((char *)pBuffer,sizeof(double)*num*2)!=sizeof(double)*num*2)
	{
		delete pBuffer;
		return false;
	}
	pointList.clear();
	for(int i=0;i<num;i++)
		pointList.append(QPointF(pBuffer[i*2],pBuffer[i*2+1]));
	delete pBuffer;
	return true;
}

bool Correspondence::writeQPointFVector(QFile &file, const QVector<QPointF> &vector)
{
	return writeQPointFList(file,QList<QPointF>::fromVector(vector));
}

bool Correspondence::readQPointFVector(QFile& file,QVector<QPointF>& vector)
{
	QList<QPointF> pointList;
	if(!readQPointFList(file,pointList))
		return false;
	vector=QVector<QPointF>::fromList(pointList);
	return true;
}
