#ifndef _Correspondence
#define _Correspondence


#include <QObject>
#include <QList>
#include <QPointF>
#include <QPainter>
#include "Image.h"
#include "Matrix.h"
#include "Vector.h"
#include "SortQueue.h"
#include <QFile>

class Correspondence: public QObject
{
	Q_OBJECT
public:
	enum MotionType{Translation,Affine,Homography,PiecewiseSmooth};
private:
	QList<QPointF> pointList1; // the correspondence between these two point sets
	QList<QPointF> pointList2;

	QList<bool> uncertaintyList; // the list of uncertainty matching, these parameters are for uncertain matching
        QList<QVector<QPointF> > candidateList;
        QList<QVector<double> > densityList;

	QList<Matrix> invCovList;
	QList<Vector> meanList;

	int pointIndex;
	MotionType motionType;
	bool IsMotionTypeSpecified;
	
	Vector transPara; //,scaledTransPara;
	Matrix affinePara; //,scaledAffinePara;
	Vector homographyPara; //,scaledHomographyPara;

	double m_Lambda; // the coefficient control how smooth the flow field should be
	DImage flow; //scaledFlow;
	bool IsFlowInterpolated;

	static int m_Left,m_Top,m_Width,m_Height;

public slots:
	void applyTransform(const QImage& image2,QImage& warpImage2,double scaleFactor);
	void specifyMotionType(int index);
signals:
	int motionTypeChanged(int index);
public:
	Correspondence(void);
	Correspondence(const Correspondence& other);
	void copyData(const Correspondence& other);
	~Correspondence(void);
	Correspondence& operator=(const Correspondence& other);
	void clear();
	static void setFlowDimension(int left,int top,int width,int height){m_Left=left;m_Top=top;m_Width=width;m_Height=height;};
	double lambda() const {return m_Lambda;};
	double& lambda() {return m_Lambda;};
	bool IsUncertain(int index) const{return uncertaintyList[index];};

	void DrawCorresponance(QPainter& painter,const QPoint& ScreenOffset,double ScaleFactor);
	void addCorrespondence(const QPointF& point1,const QPointF& point2);
	void addCorrespondence(const QPointF& point1,const SortQueue<Point>& sort);
	void addCorrespondence(const Correspondence& other);
	void computeInvCov(Matrix& invCov,Vector& mean,const QVector<QPointF>& candidates,const QVector<double> density);
	void updateSecond(const QPointF& point2);
	
	int getMotionType() const;
	bool isMotionTypeSpecified() const {return IsMotionTypeSpecified;};
	void clearMotoinType(){IsMotionTypeSpecified=false;};
	//void scaleMotion(double ScaleFactor);
	void estimateMotion();
	inline void applyMotion(double& x,double& y,int j,int i);
	
	void estimateTranslation();
	void estimateAffine();
	void estimateHomography();

	void estimateTranslationUncertain();
	void estimateAffineUncertain();
	void estimateHomographyUncertain();
	void estimatePeicewiseSmooth();
	void invalidateFlowEstimate(){IsFlowInterpolated=false;};
	bool IsParametricMotion() const{if(motionType==PiecewiseSmooth) return false;else return true;};

	bool selectPoint(const QPointF& input,double scaleFactor);
	void removePoint();
	inline int size()const {return pointList1.size();};
	double norm(const QPointF& point);

	// function to access member
	inline const QPointF& getPoint1(int index) const{return pointList1[index];};
	inline const QPointF& getPoint2(int index) const {return pointList2[index];};
	inline const QPointF& getPoint2() const{return pointList2[pointIndex];};
	
	// function to crop an image
	static void getPatch(QImage& patch,const QImage& srcImage,double x,double y,int wsize,double scale=1);

	// function to generate flow field from the motion para
	void generateFlowField(DImage& vx,DImage& vy,int left,int top,int right,int bottom);

	// functions for FILE io
	bool readCorrespondence(QFile& file);
	bool writeCorrespondence(QFile& file) const;

	static bool writeQPointFList(QFile& file,const QList<QPointF>& pointList);
	static bool readQPointFList(QFile& file,QList<QPointF>& pointList);
	static bool writeQPointFVector(QFile& file,const QVector<QPointF>& vector);
	static bool readQPointFVector(QFile& file,QVector<QPointF>& vector);
};

#endif
