#ifndef _MainWindow_h
#define _MainWindow_h


#include <QMainWindow>
#include <QScrollArea>
#include <QDialog>
#include <QSlider>
#include <QSpinBox>
#include <QHBoxLayout>
#include <QTabletEvent>
#include <QTime>

#include "ImageViewer.h"
#include "VideoLabeling.h"
#include "DepthControl.h"
#include "LayerDepth.h"
#include "LayerObject.h"
#include "Inspector.h"
#include "FlowFieldViewer.h"
#include "MagViewer.h"
#include "PaintOptions.h"
#include "StrokePainter.h"

class Application: public QApplication {
   Q_OBJECT;
public:
   Application(int &c, char **v): QApplication(c, v) {}
   bool notify(QObject *rec, QEvent *ev) {
       // cDebug() << "Called Application::notify()" << endl;
       try {
           return QApplication::notify(rec, ev);
       }
       catch (char const *str) {
           cout << "EXCEPTION: " << str << endl;
           return false;
       }
   }
};



class ImageViewer;
class VideoLabeling;
class DepthControl;
class LayerDepth;

class MainWindow: public QMainWindow
{
	Q_OBJECT
public:
	MainWindow(void);
	~MainWindow(void);

	
protected:
	void resizeEvent(QResizeEvent* event);

public slots:
        void runScript(QString filename=QString());
private slots:
    void openFolder(QString filespath=QString(),QString projectname=QString(),QString startFrame=QString(),QString endFrame=QString());
    void load(QString filename=QString());
    void importXML(QString filename=QString());
	void exportXML();
	void save();
    void saveAs(QString filename=QString(),bool useFramePath=false);
	void exportLabeling();
	void outputDepthMap();
	void outputMotion();
	void outputLayerMotion();

	void adaptScrollBar(double factor);
	void setFrameIndexRange(int nFrames);
	void setFrameControl(bool firstFrame,bool preFrame,bool nextFrame,bool lastFrame);
	void setLayerDepth(const QList<LayerDepth>& layerDepthList,int activeLayerIndex,int frameIndex);
	void reduceFrameRange();

	void playpause();
	void finishedPainting();

private:
	void createActions();
	void createMenus();
	void createToolBar();
	void binding();


	
	
	ImageViewer* imageViewer;
	QScrollArea* scrollArea;
	VideoLabeling* pVideoLabeling;

	// Menus
	QMenu* fileMenu;
	QMenu* viewMenu;
	QMenu* actionMenu;
	QMenu* windowMenu;
	QMenu* displayMenu;
	QMenu* strokeToolsMenu;
	QMenu* strokeActionsMenu;

	// Toolbars
	QToolBar* fileToolBar;
	QToolBar* zoomToolBar;
	QToolBar* strokeActionToolBar;
	QToolBar* videoToolBar;
	QToolBar* transparencyToolBar;

	// Actions for files
	QAction* openFolderAct;
	QAction* loadProjectAct;
	QAction* saveAct;
	QAction* saveAsAct;
	QAction* runScriptAct;
	QAction* importXMLAct;
	QAction* exportXMLAct;
	QAction* exportAct;
	QAction* exitAct;
	QAction* outputDepthAct;
	QAction* outputMotionAct;
	QAction* outputLayerMotionAct;

	QSpinBox* startFrameSpinBox;
	QSpinBox* endFrameSpinBox;


	// Actions for zooming the images
	QAction* zoomInAct;
	QAction* zoomOutAct;
	QAction* normalSizeAct;

	// Actions for "actions"
	QAction* reduceFrameRangeAct;

	// Actions for displaying the windows
	QAction* showInspectorAct;
	QAction* showFlowFieldViewerAct;
	QAction* hideInspectorAct;
	QAction* hideFlowFieldViewerAct;

	// Actions for displaying video elements
	QAction* showImage;
	QAction* showStrokeSpines;
	QAction* showStrokes;
	QAction* showLayerBoundaries;
	QAction* showSeedMask;
	QAction* hideImage;
	QAction* hideStrokeSpines;
	QAction* hideStrokes;
	QAction* hideLayerBoundaries;

	// the actions for displaying the video
	QAction* nextFrameAct;
	QAction* preFrameAct;
	QAction* firstFrameAct;
	QAction* lastFrameAct;
	QAction* playAct;
	//QAction* pauseAct;

	// Actions for layer operations
	QAction* createLayerAct;
	QAction* generateLayerAct;
	QAction* motionAnalysisAct;

	// Actions for stroke
	QAction* selectContourAct;
	QAction* createStrokeAct;
	QAction* overdrawContourAct;
	QAction* colorPickerAct;
	QAction* drawStrokeSelectAct;
	QAction* moveSelectedStrokesAct;
	QAction* pushSelectedStrokesAct;

	QAction *generateStrokeLayer;
	QAction *generateStrokeLayersFromList;
	QAction *deleteStrokeLayers;
	QAction *undoLastLayer;
	QAction *changeRegionBoundaryBlur;
	QAction *replicateSelectedStrokes;
	QAction *fixSelectedStrokes;
	QAction *deleteSelectedStrokes;
	QAction *deleteSelectedStrokesBackward;
	QAction *deleteSelectedStrokesForward;
	QAction *undoLastAction;
	QAction *renderAction;
	QAction *renderWithLightingAction;
	QAction *renderWithCurrentTextureAction;
	QAction *increaseStrokeWidthAction;
	QAction *decreaseStrokeWidthAction;

	QAction *showColorStrokes;
	QAction *showGuideStrokes;
	QAction *showSeedStrokes;
	QAction *showRegularStrokes;
    QTime time;

	


	//Actions for strokes

	QSpinBox* spinBoxFrameIndex;
	QSlider* sliderFrameIndex;
	
	QString framePath;
	QString projPath;
	QString projFileName;
	QString outputPathName;
	bool IsProjFileSpecified;

	DepthControl* depthControl;

	Inspector* inspector;
	FlowFieldViewer* flowFieldViewer;
	MagViewer* magViewer;
	PaintOptions* paintOptions;
	StrokePainter* strokePainter;

	bool currentlyPlaying;
};

#endif
