#ifndef _MotionPara_h
#define _MotionPara_h

#include <QList>

#include <QFile>

//-------------------------------------------------------------------------------------------
// this class defines the objective function
//-------------------------------------------------------------------------------------------
class OpticalFlowPara{
public:
	enum RegularizationType{membrane,thinplate};
private:
	bool m_IsActive;
	double m_alpha;
	double m_gamma;
	double m_eta;
	RegularizationType m_regularization;
	bool m_IsComputed;
public:
	OpticalFlowPara(){
		m_IsActive=true;
		m_alpha=0.02;
		m_gamma=0.01;
		m_eta=0.5;
		m_regularization=membrane;
		m_IsComputed=false;
	}
	OpticalFlowPara(const OpticalFlowPara& other){copyData(other);};
	void copyData(const OpticalFlowPara& other){
		m_IsActive=other.m_IsActive;
		m_alpha=other.m_alpha;
		m_gamma=other.m_gamma;
		m_eta=other.m_eta;
		m_regularization=other.m_regularization;
		m_IsComputed=other.m_IsComputed;
	}
	OpticalFlowPara& operator =(const OpticalFlowPara& other){copyData(other);return *this;};
	// function to compare whether two instances are equal
	// note that not all the fields are used for the comparison
	bool operator==(const OpticalFlowPara& other) const{
		if(m_IsActive==other.m_IsActive && m_alpha==other.m_alpha && m_gamma==other.m_gamma && m_eta==other.m_eta && m_regularization==other.m_regularization)
			return true;
		else
			return false;
	}
	OpticalFlowPara(bool isActive,double Alpha,double Gamma,double Eta,RegularizationType Regularization=membrane,bool isComputed=false)
	{
		m_IsActive=isActive;
		m_alpha=Alpha;
		m_gamma=Gamma;
		m_eta=Eta;
		m_regularization=Regularization;
		m_IsComputed=isComputed;
	}
	void display() const{
		printf("IsActive=%d  alpha=%f gamma=%f eta=%f regularization=%s IsComputed=%d\n ",m_IsActive,m_alpha,m_gamma,m_eta,m_regularization==membrane?"membrane":"thinplate",m_IsComputed);
	}
	inline bool& IsActive(){return m_IsActive;};
	inline double& alpha(){return m_alpha;};
	inline double& gamma(){return m_gamma;};
	inline double& eta(){return m_eta;};
	inline RegularizationType& regularization(){return m_regularization;};
	inline bool& IsComputed(){return m_IsComputed;};
	inline const bool IsActive()const {return m_IsActive;};
	inline const double alpha()const {return m_alpha;};
	inline const double gamma()const {return m_gamma;};
	inline const double eta() const {return m_eta;};
	inline const RegularizationType regularization()const {return m_regularization;};
	inline const bool IsComputed() const {return m_IsComputed;};

	bool writeOpticalFlowPara(QFile& file) const 
	{
		file.write((char *)&m_IsActive,sizeof(bool));
		file.write((char *)&m_alpha,sizeof(double));
		file.write((char *)&m_gamma,sizeof(double));
		file.write((char *)&m_eta,sizeof(double));
		file.write((char *)&m_regularization,sizeof(RegularizationType));
		file.write((char *)&m_IsComputed,sizeof(bool));
		return true;
	}
	bool readOpticalFlowPara(QFile& file)
	{
		file.read((char *)&m_IsActive,sizeof(bool));
		file.read((char *)&m_alpha,sizeof(double));
		file.read((char *)&m_gamma,sizeof(double));
		file.read((char *)&m_eta,sizeof(double));
		file.read((char *)&m_regularization,sizeof(RegularizationType));
		file.read((char *)&m_IsComputed,sizeof(bool));
		return true;
	}

};

//-------------------------------------------------------------------------------------------
// this class defines the parameters for the algorithm
//-------------------------------------------------------------------------------------------
class AlgorithmPara{
private:
	double m_Ratio; // the ratio for downsampling
	int m_MinWidth; // the minimum width;
	int m_nIterations; // the number of iterations;
	int m_nIRLSIterations; // the number of iterative reweighted least square (IRLS) iterations
	int m_nCGIterations; // the number of conjugate gradient iterations
public:
	AlgorithmPara(){
		/*
		m_Ratio=0.75;
		m_MinWidth=20;
		m_nIterations=16;
		m_nIRLSIterations=1;
		m_nCGIterations=50;
		*/
		m_Ratio=0.5;
		m_MinWidth=5;
		m_nIterations=5;
		m_nIRLSIterations=1;
		m_nCGIterations=10;
	}
	//AlgorithmPara(double _Ratio,int _MinWidth,int _nIterations,int _nIRLSIterations,int _nCGIterations);
	AlgorithmPara(const AlgorithmPara& other){copyData(other);};
	AlgorithmPara& operator=(const AlgorithmPara& other){copyData(other);return *this;};
	// function to compare whether two instances are equal
	bool operator==(const AlgorithmPara& other) const{
		if(m_Ratio==other.m_Ratio && m_MinWidth==other.m_MinWidth && m_nIterations==other.m_nIterations && m_nIRLSIterations==other.m_nIRLSIterations && m_nCGIterations==other.m_nCGIterations)
			return true;
		else
			return false;
	}
	void copyData(const AlgorithmPara& other){
		m_Ratio=other.m_Ratio;
		m_MinWidth=other.m_MinWidth;
		m_nIterations=other.m_nIterations;
		m_nIRLSIterations=other.m_nIRLSIterations;
		m_nCGIterations=other.m_nCGIterations;
	}
	// functions to access the members
	inline double& ratio() {return m_Ratio;};
	inline int& minWidth() {return m_MinWidth;};
	inline int& nIterations() {return m_nIterations;};
	inline int& nIRLSIterations() {return m_nIRLSIterations;};
	inline int& nCGIterations() {return m_nCGIterations;};
	inline double ratio() const {return m_Ratio;};
	inline int minWidth() const {return m_MinWidth;};
	inline int nIterations() const {return m_nIterations;};
	inline int nIRLSIterations() const {return m_nIRLSIterations;};
	inline int nCGIterations() const {return m_nCGIterations;};
	bool writeAlgorithmPara(QFile& file) const
	{
		file.write((char *)&m_Ratio,sizeof(double));
		file.write((char *)&m_MinWidth,sizeof(int));
		file.write((char *)&m_nIterations,sizeof(int));
		file.write((char *)&m_nIRLSIterations,sizeof(int));
		file.write((char *)&m_nCGIterations,sizeof(int));
		return true;
	}
	bool readAlgorithmPara(QFile& file)
	{
		file.read((char *)&m_Ratio,sizeof(double));
		file.read((char *)&m_MinWidth,sizeof(int));
		file.read((char *)&m_nIterations,sizeof(int));
		file.read((char *)&m_nIRLSIterations,sizeof(int));
		file.read((char *)&m_nCGIterations,sizeof(int));
		return true;
	}
};

//-------------------------------------------------------------------------------------------
// MotionPara is the class where all the parameters are put together to specify
// Optical Flow computation
//-------------------------------------------------------------------------------------------

class MotionPara
{
private:
	QList<OpticalFlowPara> m_FlowParaList;
	AlgorithmPara m_AlgorithmPara;
public:
	MotionPara(void);
	~MotionPara(void);
	void copyData(const MotionPara& other);
	MotionPara& operator = (const MotionPara& other);

	void addFlowPara(const OpticalFlowPara& flowPara) {m_FlowParaList.append(flowPara);};
	void removeFlowPara(int index){m_FlowParaList.removeAt(index);};
	void clear() {m_FlowParaList.clear();};

	// functions to directly access the member variables
	inline const OpticalFlowPara& flowPara(int index) const{return m_FlowParaList[index];};
	inline const QList<OpticalFlowPara>& flowParaList() const{return (const QList<OpticalFlowPara>&)m_FlowParaList;};
	inline QList<OpticalFlowPara>& flowParaList() {return (QList<OpticalFlowPara>&)m_FlowParaList;};
	inline const AlgorithmPara& algorithmPara() const {return m_AlgorithmPara;};
	inline AlgorithmPara& algorithmPara() {return m_AlgorithmPara;};

	int numFlowPara() const;

	double ratio() const{return m_AlgorithmPara.ratio();};
	int minWidth() const {return m_AlgorithmPara.minWidth();};
	int nIterations() const {return m_AlgorithmPara.nIterations();};
	int nIRLSIterations() const {return m_AlgorithmPara.nIRLSIterations();};
	int nCGIterations() const {return m_AlgorithmPara.nCGIterations();};

	double &ratio() {return m_AlgorithmPara.ratio();};
	int& minWidth() {return m_AlgorithmPara.minWidth();};
	int& nIterations() {return m_AlgorithmPara.nIterations();};
	int& nIRLSIterations() {return m_AlgorithmPara.nIRLSIterations();};
	int& nCGIterations() {return m_AlgorithmPara.nCGIterations();};

	void dispFlowPara() const;
	bool writeMotionPara(QFile& file) const;
	bool readMotionPara(QFile& file);
};

#endif