#ifndef STROKEPAINTER_H
#define STROKEPAINTER_H


#include <QGLWidget>
#include <QGLContext>
#include <QPainter>
#include "ImageViewer.h"
#include "LayerObject.h"
#include "Image.h"
#include "Stroke.h"
#include "StrokeProperties.h"
#include "StrokeLayerProperties.h"
#include "jl_vectors.h"
#include <ANN/ANN.h>
#include <QTime>
#include "emboss.h"
#include "tri_logger.hpp"

#define PI 3.1415927


//this class is for the per-frame stroke generation inputs
//it's a seperate class than the stroke layer parameters since these are the
//inputs that won't change as long as the layer and frames stay the same
class strokeGenInputs
{
public:	
	strokeGenInputs(){};
	void clear(){
		masks.clear();
		imgs.clear();
		colorStrokeImgs.clear();
		colorStrokeAlphaImgs.clear();
		visited.clear();
		edges.clear();
		seedStrokeImgs.clear();
		usrDisp.clear();
		usrPts.clear();
		usrFlowPts.clear();
		usrFlowDisp.clear();
		usrReverseFlowDisp.clear();
		usrColorPts.clear();
		usrColorDisp.clear();
		usrColorIndex.clear();
		startFrame =-1;
		endFrame = -1;
		layerName = QString("");
	};

	QVector<BinaryImage> masks;
	QVector<BiImage> imgs;
	QVector<BiImage> colorStrokeImgs;
	QVector<BiImage> colorStrokeAlphaImgs;
	QVector<BinaryImage> visited;
	QVector<BiImage> edges;
	QVector<QImage> seedStrokeImgs;
    QVector<QVector<Vec2f> > usrDisp;
    QVector<QVector<Vec2f> > usrPts;
    QVector<QVector<Vec2f> > usrFlowPts;
    QVector<QVector<Vec2f> > usrFlowDisp;
    QVector<QVector<Vec2f> > usrReverseFlowDisp;
    QVector<QVector<Vec2f> > usrColorPts;
    QVector<QVector<int> > usrColorIndex;
    QVector<QVector<Vec2f> > usrColorDisp;

	int startFrame;
	int endFrame;
	QString layerName;

};

//Info on the 
class StrokeGenInfo
{
public:	
	StrokeGenInfo(){clear();};
	void clear(){
		count=0;
		tooShort=0;
		thrownOutside=0;
		outsideFlowImage=0;
		lastFrameTooShort=0;
		boundaryClip=0;
		strokeOutsideBox=0;
		outsideMask=0;
		pastMagnitude = 0;
		pastDensity = 0;
		colorThresh = 0;
		brightnessThresh = 0;
		madeToEnd = 0;
		addedStrokes = 0;
		guideStrokeIntersection=0;
		directionError =0;
		strokeOptimizationThreshold=0;
	};
	void print()
	{

		LOG_COUT(count);
		LOG_COUT(tooShort);
		LOG_COUT(thrownOutside);
		LOG_COUT(outsideFlowImage);
		LOG_COUT(lastFrameTooShort);
		LOG_COUT(strokeOutsideBox);
		LOG_COUT(outsideMask);
		LOG_COUT(boundaryClip);
		LOG_COUT(pastMagnitude);
		LOG_COUT(pastDensity);
		LOG_COUT(colorThresh);
		LOG_COUT(brightnessThresh);
		LOG_COUT(madeToEnd);
		LOG_COUT(addedStrokes);
		LOG_COUT(guideStrokeIntersection);
		LOG_COUT(directionError);
		LOG_COUT(strokeOptimizationThreshold);
		LOG_COUT((float)addedStrokes/(float)madeToEnd);
	}

	int count;
	int tooShort;
	int thrownOutside;
	int outsideFlowImage;
	int lastFrameTooShort;
	int strokeOutsideBox;
	int outsideMask;
	int boundaryClip;
	int pastMagnitude;
	int pastDensity;
	int colorThresh;
	int brightnessThresh;
	int madeToEnd;
	int addedStrokes;
	int guideStrokeIntersection;
	int directionError;
	int strokeOptimizationThreshold;

};


class StrokePainter : public QGLWidget
{
	Q_OBJECT

public:
	StrokePainter(QWidget *parent = 0, int width=0, int height=0,int frames=1);
	StrokePainter(const StrokePainter& strokePainter);
	~StrokePainter();

	void clear();

	void setSize(int width, int height);
	void gotoFrame(int frame);
	//void applyBoundaryBlur(QImage& img,const BinaryImage& mask, const QRect& rect,double layerBoundaryBlur);
	void applyMedianFilter(QImage& img,double layerMedianBlur);

signals:
	void strokePropertiesChanged(StrokeProperties& sp);
	void strokePropertiesChanged(QString prop, QString val);

	void generatingStrokes(bool);

public slots:

	//setting the current stroke properties from the paint options widget
	void loadTextureForCurrentStroke(QString& filename);
	void strokeStartWidthChanged(int startWidth);
	void strokeEndWidthChanged(int endWidth);
	void guideStroke(bool isGuide);
	void useImageColor(bool useImageColor);
	void useColor(QColor col);
	void alphaChanged(double alpha);
	//void colorVarianceChanged(int colorVariance);

	void loadPaperTexture(QString& filename);
	void usePaperColor(QColor col);

	void hueOffsetChanged(int hueOffset){printf("hueOffsetChanged %i\n",hueOffset);_interfaceSlp.hueOffset = hueOffset;};
	void saturationOffsetChanged(int saturationOffset){printf("saturationOffsetChanged %i\n",saturationOffset);_interfaceSlp.saturationOffset = saturationOffset;};
	void valueOffsetChanged(int valueOffset){printf("valueOffsetChanged %i\n",valueOffset);_interfaceSlp.valueOffset = valueOffset;};
	void hueNoiseChanged(double hueNoise){printf("hueNoiseChanged %f\n",hueNoise);_interfaceSlp.hueNoise = hueNoise;};
	void saturationNoiseChanged(double saturationNoise){printf("saturationNoise %f\n",saturationNoise);_interfaceSlp.saturationNoise = saturationNoise;};
	void valueNoiseChanged(double valueNoise){printf("valueNoiseChanged %f\n",valueNoise);_interfaceSlp.valueNoise = valueNoise;};

	//color strokes
	void colorStroke(bool);
	void visibleStroke(bool);
	void fillStroke(bool);
	void seedStroke(bool);
	void addVisibleStrokesToLayer(bool addVisibleStrokesToLayer){_interfaceSlp.addVisibleStrokesToLayer=addVisibleStrokesToLayer;};

	//void passOverGuideStrokes(bool passOverGuideStrokes){_interfaceSlp.passOverGuideStrokes=passOverGuideStrokes;};
	//void darkenOccludedRegionsChanged(bool darkenOccludedRegions){_interfaceSlp.darkenOccludedRegions=darkenOccludedRegions;};
	void shiftToEdgesChanged(bool shiftToEdges){_interfaceSlp.shiftToEdges=shiftToEdges;printf("shiftToEdges %i\n",_interfaceSlp.shiftToEdges);};
	void optimizeStrokesChanged(bool optimizeStrokes){_interfaceSlp.optimizeStrokes=optimizeStrokes;printf("optimizeStrokes %i\n",_interfaceSlp.optimizeStrokes);};
	void useStrokesFromAllLayersChanged(bool useStrokesFromAllLayers){_interfaceSlp.useStrokesFromAllLayers=useStrokesFromAllLayers;printf("useStrokesFromAllLayers %i\n",_interfaceSlp.useStrokesFromAllLayers);};


	void seedOrderChanged( const QString &seedOrder){_interfaceSlp.seedOrder=seedOrder;};

	void optimizeStrokeWidthsChanged(bool optimizeStrokeWidths){printf("optimizeStrokeWidths %i\n",optimizeStrokeWidths);_interfaceSlp.optimizeStrokeWidths = optimizeStrokeWidths;};
	void oswColorFactorChanged(double oswColorFactor){printf("oswColorFactor %f\n",oswColorFactor);_interfaceSlp.oswColorFactor = oswColorFactor;};
	void oswCoherenceFactorChanged(double oswCoherenceFactor){printf("oswCoherenceFactor %f\n",oswCoherenceFactor);_interfaceSlp.oswCoherenceFactor = oswCoherenceFactor;};
	void oswRegularizationFactorChanged(double oswRegularizationFactor){printf("oswRegularizationFactor %i\n",oswRegularizationFactor);_interfaceSlp.oswRegularizationFactor = oswRegularizationFactor;};
	void oswGrowFactorChanged(double oswGrowFactor){printf("oswGrowFactor %f\n",oswGrowFactor);_interfaceSlp.oswGrowFactor = oswGrowFactor;};
	void oswEdgeFactorChanged(double oswEdgeFactor){printf("oswEdgeFactor %f\n",oswEdgeFactor);_interfaceSlp.oswEdgeFactor = oswEdgeFactor;};

	void renderFrames(){_renderFrames = true;};
	void renderScaleChanged(double renderScale){_renderScale = renderScale;printf("renderScale %f\n",_renderScale);};
	void renderWithLighting(bool renderWithLighting){_renderWithLighting = renderWithLighting;printf("render with lighting %i\n", _renderWithLighting);};
	void renderWithCurrentTexture(bool renderWithCurrentTexture){_renderWithCurrentTexture=renderWithCurrentTexture;printf("renderWithCurrentTexture %i\n", renderWithCurrentTexture);};

	void useBoundaryFlow(bool useBoundaryFlow){printf("useBoundaryFlow %i\n", useBoundaryFlow);_useBoundaryFlow = useBoundaryFlow;};
	void useStrokeFlow(bool useStrokeFlow){printf("useStrokeFlow %i\n", useStrokeFlow);_useStrokeFlow = useStrokeFlow;};
	void oldBlend(bool oldBlend){printf("oldBlend %i\n", oldBlend);_oldBlend = oldBlend;};
	void multiPassRender(bool multipassrender){_multiPassRender = multipassrender;};

	void drawPaperAndImage();

	
	//void freeTexture(int texture){if (texture>1)glDeleteTextures( 1, &texture );};

	
	//setting the filling parameters from the paint options widget
	void orientationChanged(int orientation){_interfaceSlp.orientationAngle = ((double)(360+(-1*orientation-90)%360) * (PI/180.0));printf("orientation(radians) %f\n",_interfaceSlp.orientationAngle);};
	void includeReverseDirection(bool reverseDirection){_interfaceSlp.reverseDirection = reverseDirection;};
	void fixedOrientationFactorChanged(int fixedOrientationFactor){_interfaceSlp.fixedOrientationFactor = fixedOrientationFactor;};
	void followBoundaryFactorChanged(int boundaryFactor){_interfaceSlp.boundaryFactor = boundaryFactor;};
	void followImageFactorChanged(int imageGradientFactor){_interfaceSlp.imageGradientFactor = imageGradientFactor;};
	void followUserStrokesFactorChanged(int userStrokesFactor){_interfaceSlp.userStrokesFactor = userStrokesFactor;printf("userStrokesFactor %i\n",_interfaceSlp.userStrokesFactor); };
	void orientationVarianceChanged(double orientationVariance){_interfaceSlp.orientationVariance = orientationVariance;printf("variance %f\n",_interfaceSlp.orientationVariance);};
	void paintOccludedRegionsChanged(bool paintOccludedRegions){_interfaceSlp.paintOccludedRegions = paintOccludedRegions;printf("paint occluded %i\n",_interfaceSlp.paintOccludedRegions); };
	void startInOccludedRegionsChanged(bool startInOccludedRegions){_interfaceSlp.startInOccludedRegions = startInOccludedRegions;printf("start in occluded %i\n",_interfaceSlp.startInOccludedRegions); };
	
	void useOccludingLayerFlowChanged(bool useOccludingLayerFlow){_interfaceSlp.useOccludingLayerFlow = useOccludingLayerFlow;printf("useOccludingLayerFlow %i\n",_interfaceSlp.useOccludingLayerFlow); };
	void clipToSeedRegionsChanged(bool clipToSeedRegions){_interfaceSlp.clipToSeedRegions = clipToSeedRegions;printf("clipToSeedRegions %i\n",clipToSeedRegions); };
	

	

	void fillEmptyRegionsChanged(bool fillEmptyRegions){_interfaceSlp.fillEmptyRegions = fillEmptyRegions;printf("fillEmptyRegions %i\n",fillEmptyRegions); };
	void fillColorDifferenceChanged(bool fillColorDifference){_interfaceSlp.fillColorDifference = fillColorDifference;printf("fillColorDifference %i \n",fillColorDifference); };
	void fillUserSeedChanged(bool fillUserSeed){_interfaceSlp.fillUserSeed = fillUserSeed;printf("_interfaceSlp.fillUserSeed %i \n",_interfaceSlp.fillUserSeed); };
	
	void startHeavySampleChanged(int startHeavySample){printf("startHeavySample %i\n",startHeavySample); _startHeavySample = startHeavySample;};
	void endHeavySampleChanged(int endHeavySample){printf("endHeavySample %i\n",endHeavySample); _endHeavySample = endHeavySample;};

	void minStrokeLengthChanged(int minStrokeLength){_interfaceSlp.minStrokeLength= minStrokeLength;};
	void maxStrokeLengthChanged(int maxStrokeLength){_interfaceSlp.maxStrokeLength = maxStrokeLength;};
    void strokeDensityChanged(int strokeDensity){_interfaceSlp.strokeDensity = ((float)strokeDensity)/1000.;printf("stroke density %f %i \n",_interfaceSlp.strokeDensity, strokeDensity);};
	void densityWidthAddChanged(double densityWidthAdd){_interfaceSlp.densityWidthAdd = densityWidthAdd;printf("densityWidthAdd %f \n",densityWidthAdd);};
	void densityBrightnessFactorChanged(double densityBrightnessFactor){_interfaceSlp.densityBrightnessFactor= densityBrightnessFactor;printf("densityBrightnessFactor %f \n",densityBrightnessFactor);};
	
	void minBrightnessChanged(int minBrightness){_interfaceSlp.minBrightness = minBrightness;printf("minBrightness %i \n",minBrightness);};
	void maxBrightnessChanged(int maxBrightness){_interfaceSlp.maxBrightness = maxBrightness;printf("maxBrightness %i \n",maxBrightness);};
	

	void magnitudeThresholdChanged(int magnitudeThreshold){_interfaceSlp.magnitudeThreshold = (float)magnitudeThreshold;};

	void colorDifferenceChanged(double colorDifference){_interfaceSlp.colorDifference = colorDifference;printf("colorDifference %f\n", _interfaceSlp.colorDifference);};
	void maxStrokeAngleChanged(int maxStrokeAngle){printf("maxAngle %i\n", maxStrokeAngle);_interfaceSlp.maxStrokeAngle = maxStrokeAngle;}
	void radiusFactorChanged(double radiusFactor){printf("radiusFactor %f\n", radiusFactor);_interfaceSlp.radiusFactor = radiusFactor;}

	void minFramesChanged(int minFrames){printf("min frames %i\n", minFrames);_interfaceSlp.minFrames = minFrames;}
	void maxFramesChanged(int maxFrames){printf("max frames %i\n", maxFrames);_interfaceSlp.maxFrames = maxFrames;}
	void maxStrokesChanged(int maxStrokes){printf("max strokes %i\n", maxStrokes);_interfaceSlp.maxStrokes = maxStrokes;}
	void requestedStartFrameChanged(int requestedStartFrame){_interfaceSlp.requestedStartFrame = requestedStartFrame;};
	void requestedEndFrameChanged(int requestedEndFrame){_interfaceSlp.requestedEndFrame = requestedEndFrame;};
	void colorUpdateChanged(double colorUpdate){printf("color update %f\n", colorUpdate);_interfaceSlp.colorUpdate = colorUpdate;};
	void fadeRateChanged(int fadeRate){printf("fadeRate %i\n", fadeRate);_interfaceSlp.fadeRate = fadeRate;};
	void followStrokesFactorChanged(double followStrokesFactor){printf("followStrokesFactor %f\n", followStrokesFactor);_interfaceSlp.followPreviousStrokesFactor = followStrokesFactor;};
	void followGradientFactorChanged(double followGradientFactor){printf("followGradientFactor %f\n", followGradientFactor);_interfaceSlp.followGradientFactor = followGradientFactor;};

	void coherenceFactorChanged(double coherenceFactor){printf("coherenceFactor %f\n", coherenceFactor);_interfaceSlp.coherenceFactor = coherenceFactor;};
	void orientationFactorChanged(double orientationFactor){printf("orientationFactor %f\n", orientationFactor);_interfaceSlp.orientationFactor = orientationFactor;};
	void colorFactorChanged(double colorFactor){printf("colorFactor %f\n", colorFactor);_interfaceSlp.colorFactor = colorFactor;};
	void densityFactorChanged(double densityFactor){printf("densityFactor %f\n", densityFactor);_interfaceSlp.densityFactor = densityFactor;};
	void rigidityFactorChanged(double rigidityFactor){printf("rigidityFactor %f\n", rigidityFactor);_interfaceSlp.rigidityFactor = rigidityFactor;};
	


	void systemStrokeTransparencyChanged(int strokeTransparency){_systemStrokeTransparency = ((float)strokeTransparency)/100.0;_redraw=true;printf("systemStrokeTransparency %i\n",strokeTransparency);};
	void userStrokeTransparencyChanged(int strokeTransparency){_userStrokeTransparency = ((float)strokeTransparency)/100.0;_redraw=true;printf("userStrokeTransparency %i\n",strokeTransparency);};
	void imageTransparencyChanged(int imageTransparency){_imageTransparency = ((float)imageTransparency)/100.0;_redraw=true;printf("imageTransparency %i\n",imageTransparency); };
	

	void userFlowFactorChanged(int userFlowFactor){printf("userFlowFactor %i\n", userFlowFactor);_interfaceSlp.userFlowFactor = userFlowFactor;};
	void imageFlowFactorChanged(int imageFlowFactor){printf("imageFlowFactor %i\n", imageFlowFactor);_interfaceSlp.imageFlowFactor = imageFlowFactor;};
	void boundaryFlowFactorChanged(int boundaryFlowFactor){printf("boundaryFlowFactor %i\n", boundaryFlowFactor);_interfaceSlp.boundaryFlowFactor = boundaryFlowFactor;}
	void fixedFlowFactorChanged(int fixedFlowFactor){printf("fixedFlowFactor %i\n", fixedFlowFactor);_interfaceSlp.fixedFlowFactor = fixedFlowFactor;}
	

	void maxFlowAngleChanged(int maxFlowAngle){printf("maxFlowAngle %i\n", maxFlowAngle);_interfaceSlp.maxFlowAngle = maxFlowAngle;}
	void fixedMovementDirectionChanged(int fixedMovementAngle){_interfaceSlp.fixedMovementAngle = ((double)(360+(-1*fixedMovementAngle-90)%360) * (PI/180.0));printf("fixed movement direction(radians) %f\n",_interfaceSlp.fixedMovementAngle);};
	void fixedMovementRateChanged(int fixedMovementRate){printf("fixedMovementRate %i\n", fixedMovementRate);_interfaceSlp.fixedMovementRate = fixedMovementRate;};
	void movementVarianceChanged(double movementVariance){_interfaceSlp.movementVariance = movementVariance;printf("move variance %f\n",_interfaceSlp.movementVariance);};


	//slots for signals from the main video labelling widget
	void redraw();
	void strokesChanged(int curFrameIndex);
	void numFramesChanged(int numFrames);
	void frameChanged(const QImage* m_Image,const QImage* m_NextImage);
	void Paint(ImageViewer& imageViewer,QPainter& painter);
	void paintStrokes(ImageViewer& imageViewer,QPainter& painter,QList<QImage>& imgList,QList<LayerObject>& layerObjectList,int curFrameIndex);
    void paintFrameStrokes(ImageViewer& imageViewer,QPainter& painter, QList<LayerObject>& layerObjectList,int curFrameIndex);
	void paintFrameStrokesFinal(QGLFramebufferObject *fbo,QList<LayerObject>& layerObjectList,int curFrameIndex,float renderScale,QString outputDir, bool doubleExposure,QImage &lastImg,QImage &finalImg);
	void generateStrokes(QString& path, LayerObject& layer, QList<LayerObject>& allLayers,  QList<QImage>& imgList,QStringList& fileList,int curFrameIndex,int generateId,StrokeLayerProperties *slp,bool fromScript);
	void renderStrokes(int requestedStartFrame, int requestedEndFrame, float renderScale, QList<LayerObject>& allLayers,  QList<QImage>& imgList, QString outputDir);
	

	GLuint loadStrokeImageTexture(const QImage& img);
	void setCurrentColor(QColor& color, float pointx, float pointy);
	void setCurrentColorForFilledStroke(Stroke *stroke, QImage& img);


protected:
	void initializeGL();
	void paintGL();
	void resizeGL(int width, int height, double scaleFactor);
//	void mousePressEvent(QMouseEvent *event);
//	void mouseMoveEvent(QMouseEvent *event);

private:
	int _w;
	int _h;
	int _frames;

	int _startingCheck;
	bool _testingMode;


	bool _useBoundaryFlow;
	bool _useStrokeFlow;
	bool _oldBlend;
	bool _multiPassRender;

	int _startHeavySample, _endHeavySample;

	

	//rendering parameters

	bool _renderFrames;
	bool _renderWithLighting;
	double _renderScale;


	typedef enum {CONSTANT_DEPTH = 0, MULTIPLY_DEPTH=1, ADD_DEPTH = 2} depthT;
	depthT _depthBlending;

	float _systemStrokeTransparency;
	float _userStrokeTransparency;
	float _imageTransparency;


	//paper 
	GLuint _paperTexture;
	QColor _paperColor;
	float _paperWidthRatio, _paperHeightRatio;


	//stroke drawing 
	
	bool _redraw;

	QVector<QString> _textureNames;
	QMap<int, int> _alphaTextureMap;
	QMap<int, float> _textureWidthMap;

        QMap<int, std::pair<double *,int> > _gaussianMasks;

	QVector<GLuint>_frameTextures;
	GLuint _mainTex, _imageTex,_secondaryTex;
	QGLFramebufferObject *_fboVar, *_fbo;
	double _fboscaleFactor;


	//Stroke generation inputs

	//inputs that may stay fixed
	strokeGenInputs inp;

	QVector<BiImage> _etfAngle;
	QVector<BiImage> _etfMag;
	QVector<bool> _etfCreated;

    unsigned char *_imgDensity;

	double *_gaussianMask;
	int _gaussianMaskSize;

	int _kdTreeK;


	float _magMax;

	bool _feedForwardFade;

	bool _drawSpines;


    QTime t;
    bool _readPixelsTiming;

	StrokeLayerProperties _slp;
	StrokeLayerProperties _interfaceSlp;

	StrokeProperties _currentProperties;

	bool _renderWithCurrentTexture;

	//deprecated
	int _followFlowFactor;


	char _strbuf[2048];
	GLenum _errCode;

	float _optimizationThreshold;

	void deleteFrameTextures();
	void initializePainter(const QImage& currImage);

	//bool createFrameBufferObjects();
	void multiPassStrokeRender(Stroke* s);

	void createETFImage(int frame, QList<QImage>& imgList,QStringList& fileList,QString& path);

	bool getInputs(LayerObject& layer, QList<QImage>& imgList,QStringList& fileList,QString& path, int startFrame, int endFrame);
	void getUserStrokeInputs( QList<LayerObject>& allLayers, LayerObject& layer, QList<QImage>& imgList,QString& path,int startFrame, int endFrame,bool *getUserStrokeInputs);


	int loadTexture(QString& filename);
	void drawTexture(LayerObject& layer, int frame,bool useTransparency);
	bool fillLayerWithStrokes(LayerObject& layer,QList<LayerObject>& allLayers, QList<QImage>& imgList, int generateId);
	//bool feedForwardFillLayerWithStrokes(LayerObject& layer,QList<LayerObject>& allLayers, QList<QImage>& imgList, int generateId);
	
	//void startStroke(int initx, int inity,int frame,QVector<QVector<QList<int>>>& pointQueueX, QVector<QVector<QList<int>>>& pointQueueY, LayerObject& layer,QVector<Stroke*>& createdStrokes,strokeGenInputs& inp);	
	void startStroke(int initx, int inity,int frame, LayerObject& layer,QList<LayerObject>& allLayers, QVector<Stroke*>& createdStrokes, StrokeGenInfo& sgi,QVector<QVector<pair<int,pair<int,int> > > > pointQueue,long *pointQueueSize);
	//Stroke* startStrokeFeedForward(int initx, int inity, Stroke* startStroke,int frame, LayerObject& layer,QList<LayerObject>& allLayers, QVector<Stroke*>& createdStrokes, StrokeGenInfo& sgi,QVector<QVector<pair<int,pair<int,int> > > > pointQueue,long *pointQueueSize);
	bool getNewStrokeColor(LayerObject& layer, int currFrameRel,QColor& color,QColor& imgColor,  Stroke *stroke, Vec3i& strokeNoise); //,QColor firstColor=QColor(0,0,0)
	bool getCombinedImageAndUserStrokeColor(int currFrameRel,float x, float y,double* mask, int maskSize, LayerObject& layer, float* red,float* green,float* blue);

	bool ageFade(Stroke *stroke,int frame);

	Vec2f getGreedyDirection(Vec2f& oldDir,Vec2f& lastDir, Vec2f& optFlowDir, float x, float y, int currFrameRel,int radius,int dir,float directionSwitch,LayerObject& layer, float *mag);
	Vec2f getDirection(Vec2f& oldDir,Vec2f& lastDir, Vec2f& optFlowDir, float x, float y, int currFrameRel,int radius,int dir,float directionSwitch,LayerObject& layer, float *mag,bool invMag=true);
	
	void truncateDirection(Vec2f& newDir, Vec2f& oldDir, float thresholdAngle);

	bool getFlowVector(float startx, float starty,  int frameRel,bool reverse,int startFrame, LayerObject& layer,QList<LayerObject>& allLayers, float* flowx, float* flowy);
	bool getOpticalFlow(int frame, int x, int y,LayerObject& layer,bool reverse,bool useOccludingLayer,QList<LayerObject>& allLayers,  float *vx, float *vy);
	Vec2f getGradientDirection(Vec2f& oldDir,Vec2f& lastFrameDir, float startx, float starty, int frame,int startFrame, int radius,int dir,float directionSwitch);
	Vec2f getBoundaryDirection(float startx, float starty, int relativeFrame,int startFrame, int radius,int dir,float directionSwitch,LayerObject& layer,float *dist);

	//void getImageColor(int frame, int startx, int starty, double* gaussfilter, int fsize,const BinaryImage& layerMask, float* red,float* green,float* blue );
	bool getImageColor(BiImage& img , int startx, int starty,double* gaussfilter, int fsize,const BinaryImage& layerMask, float* red,float* green,float* blue, float *grey );
	//void generateStrokeLayer(QString& path,QString& filename, LayerObject& layer,  QList<QImage>& imgList,int curFrameIndex);
	//float getWinnemollerDoGMagnitude(BiImage& img , int x, int y);

	void blendStrokeColors(QVector<Stroke*>& createdStrokes, int seedFrame);
	double *getGaussianFilter(float sigma,int *filterSize);
	bool shiftStrokeToOrientationField(int initx,int inity,int currFrameRel, float strokeRadius,LayerObject& layer, int *shiftedx,int *shiftedy);

	float getColorDist(QColor& oldcol, QColor& newcol);
	bool optimizePosition(int currFrameRel, float x,float y,QColor& startColor, float radius,Vec2f& newDir,float angleThreshold, float *shiftedx,float *shiftedy);

    float evaluateControlPoint(LayerObject& layer,Vec2f& oldDir,Vec2f& lastFrameDir,float startx, float starty, float newx, float newy, int currFrameRel, double* gaussianMask,int maskSize,float radius, int dir,int directionSwitch,QColor& startColor);
    float evaluateOrientationFieldComponent(LayerObject& layer,float startx, float starty,int currFrame,Vec2f& dir, float radius);
	float evaluateOrientationFieldComponentNew(LayerObject& layer,float startx, float starty,int currFrame,Vec2f& dir, float radius);
	
	Vec2f getUserStrokeDirection( LayerObject& layer,float x, float y,int currFrame,Vec2f& checkDir, float *mag);

	bool getNewOptimizedDirection(Vec2f& oldDir,Vec2f& lastFrameDir, float x, float y, int currFrameRel,int radius,int dir,float directionSwitch,QColor& startColor,LayerObject& layer, Vec2f& startingDir,double* gaussianMask,int maskSize, float *mag,Vec2f& optimizedDir);

	//inline float angleTransform(int angle){float ang=((float)angle)/255.; if (ang>=0.5) ang=ang-0.5; return (ang*2*PI -PI);};

	//float angleTransform(int angle);
	float angleTransform(int angle){return ((((float)angle)/255.)*2*PI) -PI;};

	float getColorDifference(int currFrameRel, int startx, int starty, double* gaussfilter, int fsize,QColor& color);


	double evaluateStrokeWidths(LayerObject& layer,int currFrameRel,float currStartWidth,float currEndWidth,float lastStartWidth,float lastEndWidth,QVector<Vec2f>& strokePoints, QColor& initCol, bool firstFrame,float curMin=999999);
	Vec2f optimizeStrokeWidth(int currFrameRel,LayerObject& layer,bool firstFrame,QVector<Vec2f>& currentStroke, float previousStartWidth, float previousEndWidth,QColor& initColor, bool firstStroke);
	void optimizeStrokeWidths(LayerObject& layer,QVector<Stroke*>& createdStrokes, Vec3i& strokeNoise);
	void getStrokeWidthDerivative(int currFrameRel,float currStartWidth,float currEndWidth,float lastStartWidth,float lastEndWidth,Stroke* stroke,bool firstStroke,double *dStartWidth,double *dEndWidth);
	void getEdgeness(BiImage& img , int startx, int starty, double* gaussfilter, int fsize, float* edgeness );

	int truncateColor(float col){ if (col >= 254.5) return 255; else if  (col<0.5) return 0; else return col;}; 
};

#endif
